/** mod_tidy.c -- a TidyLib based DSO module for the Apache HTTP Server Version 2

  COPYRIGHT NOTICE:
  
    Copyright [2002] [Sebastian Tusk]
    Copyright [2005] [Sierk Bornemann]

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

  For detailled information about this licence, please see the file called LICENCE.
   
  Contributing Author(s):

     Sebastian Tusk (Initial author)
     Sierk Bornemann <bornemann@users.sourceforge.net>

  The contributing author(s) would like to thank all those who
  helped with testing, bug fixes and suggestions for improvements. 
  This wouldn't have been possible without your help.

  Created 2002 by Sebastian Tusk
  Updated 2005-07-14 by Sierk Bornemann


  SVN Info:

    $Author: bornemann $ 
    $Date: 2006-01-26 16:44:29 +0100 (Thu, 26 Jan 2006) $ 
    $Revision: 111 $ 

*/
 
// #define MOD_TIDY_DEBUG

/* Apache headers */
#include "apr_buckets.h"
#include "apr_strings.h"
#include "apr_version.h"
#include "ap_config.h"
#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"

/* TidyLib headers */
#include "tidy.h"
#include "buffio.h"

/* mod_tidy headers */
#include "mod_tidy.h"

/* points to the Tidy Configuration Options (Quick Reference) */
#define TIDY_QUICKREF_URL "http://tidy.sourceforge.net/docs/quickref.html"

/* apr_thread_mutex_t *mod_tidy_g_MemMutex = NULL; */
apr_pool_t *mod_tidy_g_MemPool = NULL;

/* mod_tidy memory management */
void TIDY_CALL mod_tidy_FatalError( ctmbstr msg )
{
	fprintf( stderr, "Fatal error: %s\n", msg );
	exit(2);
}

void* TIDY_CALL mod_tidy_MemAlloc( size_t size )
{
	void *p;

	p = apr_pcalloc(mod_tidy_g_MemPool, size + sizeof( size_t ) );
	
	if ( p ) {
		*(size_t*)p = size;
    	p = (size_t*)p + 1;
	} else {
        mod_tidy_FatalError("Out of memory!");
	}
    return p;
}

void TIDY_CALL mod_tidy_MemFree( void* mem )
{
}

void* TIDY_CALL mod_tidy_MemRealloc( void* mem, size_t newsize )
{
	void *p;
	int oldsize;

	if (mem == (void *)NULL)
		return mod_tidy_MemAlloc(newsize);
		
	if (newsize == 0) {
		mod_tidy_MemFree( mem );
		return NULL;
	}
	
	oldsize = *((size_t*)mem - 1);

	if (oldsize >= newsize) {
		return mem;
	}
	
	p = mod_tidy_MemAlloc( newsize );
	if (!p)
		mod_tidy_FatalError("Out of memory!");
	else{
		memcpy( p, mem, oldsize );
		
		mod_tidy_MemFree( mem );
	}
    return p;
}

/* apply options from table to tidy doc */
static int mod_tidy_applyoptions(void *rec, const char *key, 
	const char *value)
{
	mod_tidy_struct *ctx = rec;
	
#ifdef MOD_TIDY_DEBUG
	printf( "mod_tidy_applyoptions\n" );
	printf( "key == %s\n", key );
	printf( "value == %s\n", value );
#endif /* #ifdef MOD_TIDY_DEBUG */
	
	if (!tidyOptParseValue(ctx->tidyDoc, key, value )) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, ctx->f->r->server,
			"error occured 'TidyOption %s %s'", key, value );
	}
	
	return TRUE;
}

static int mod_tidy_filter(ap_filter_t *f, apr_bucket_brigade *bb)
{
    mod_tidy_struct *ctx = f->ctx;
    apr_bucket *e;
    /* apr_pool_t *p = f->r->pool; */
    request_rec *r = f->r;

#ifdef MOD_TIDY_DEBUG
	printf( ">>>mod_tidy_filter( 0x%X, 0x%X )\n", f, bb );
	printf( "ctx == 0x%X\n", ctx );
#endif /* #ifdef MOD_TIDY_DEBUG */

	/* apr_thread_mutex_lock( mod_tidy_g_MemMutex ); */
	mod_tidy_g_MemPool = f->r->pool;
	
	if (ctx == NULL) {
		mod_tidy_config *cfg = ap_get_module_config(
			f->r->per_dir_config, &tidy_module);

		f->ctx = ctx = apr_pcalloc(f->r->pool, sizeof(*ctx));
		ctx->ba = apr_bucket_alloc_create(f->r->pool);
		ctx->bb = apr_brigade_create(f->r->pool, ctx->ba);
		ctx->f = f;
		
		tidySetMallocCall( mod_tidy_MemAlloc );
		tidySetReallocCall( mod_tidy_MemRealloc );
		tidySetFreeCall( mod_tidy_MemFree );
		tidySetPanicCall( mod_tidy_FatalError );

		ctx->tidyDoc = tidyCreate();
		memset( &(ctx->errbuf), 0, sizeof( ctx->errbuf ) );
		memset( &(ctx->htmlbuf), 0, sizeof( ctx->htmlbuf ) );
		
		tidySetErrorBuffer( ctx->tidyDoc, &(ctx->errbuf) );
		
		/* set tidy options; iterate all options */
		apr_table_do( mod_tidy_applyoptions, ctx, 
			cfg->tidyoptions, NULL );
	}

	for (e = APR_BRIGADE_FIRST(bb);
		e != APR_BRIGADE_SENTINEL(bb);
		e = APR_BUCKET_NEXT(e)) {
		const char *str;
		apr_size_t len;

		if (APR_BUCKET_IS_EOS(e)) { /* * || APR_BUCKET_IS_FLUSH(e) */
			int errors = 0;
			int line = 1;
			TidyBuffer tidyhtmlbuf;
		
			tidyParseBuffer( ctx->tidyDoc, &ctx->htmlbuf );
			ctx->htmlbuf.next = 0;
			
			memset( &(tidyhtmlbuf), 0, sizeof( tidyhtmlbuf ) );
			tidySaveBuffer( ctx->tidyDoc, &tidyhtmlbuf );
			
			if (tidyErrorCount(ctx->tidyDoc) || tidyWarningCount(ctx->tidyDoc) ) {
				errors = 1;
			}
		
			if (errors) {
				char strline[6];
				TidyBuffer escaped_errbuf;
				
				/* tidyBufAppend( &ctx->errbuf, "\n\n", 2 ); */
				tidyErrorSummary( ctx->tidyDoc );
				
				/* replace & < > with named entities */
				memset( &(escaped_errbuf), 0, sizeof( escaped_errbuf ) );
				
				while( !tidyBufEndOfInput( &ctx->errbuf ) ) {
					int ch = tidyBufGetByte( &ctx->errbuf );
					
					if (ch == '&') {
						tidyBufAppend( &escaped_errbuf, "&amp;", 5 );
					}else if (ch == '<') {
						tidyBufAppend( &escaped_errbuf, "&lt;", 4 );
					}else if (ch == '>') {
						tidyBufAppend( &escaped_errbuf, "&gt;", 4 );
					}else if (ch == '\n') {
						tidyBufAppend( &escaped_errbuf, "</li>\n<li class=\"msg_err\">", 26 );
					}else{
						tidyBufPutByte( &escaped_errbuf, ch );
					}
				}

				tidyBufAppend( &escaped_errbuf, "</li>\n</ol>\n</div>\n", 19 );
				tidyBufAppend( &escaped_errbuf, "<div id=\"source\">\n", 18 );
				tidyBufAppend( &escaped_errbuf, "<h2>Source Listing</h2>\n", 24 );
				tidyBufAppend( &escaped_errbuf, "<p>Below is the source input I used for checking:</p>\n", 54 );
				tidyBufAppend( &escaped_errbuf, "<ul>\n<li><code>", 15 );

				tidyBufAppend( &escaped_errbuf, "0001 ", 5 );
				line = 1;
				
				while( !tidyBufEndOfInput( &ctx->htmlbuf ) ) {
					int ch = tidyBufGetByte( &ctx->htmlbuf );
					
					if (ch == '&') {
						tidyBufAppend( &escaped_errbuf, "&amp;", 5 );
					}else if (ch == '<') {
						tidyBufAppend( &escaped_errbuf, "&lt;", 4 );
					}else if (ch == '>') {
						tidyBufAppend( &escaped_errbuf, "&gt;", 4 );
					}else if (ch == '\n') {
						tidyBufAppend( &escaped_errbuf, "</code></li>\n<li><code>", 23 );
						line ++;
						if (line < 1000) {
							sprintf( strline, "%.4d ", line );
							tidyBufAppend( &escaped_errbuf, strline, 5 );
						}
					}else{
						tidyBufPutByte( &escaped_errbuf, ch );
					}
				}

				tidyBufAppend( &escaped_errbuf, "</code></li>\n</ul></div>\n", 25 );
				tidyBufAppend( &escaped_errbuf, "<div id=\"output\">\n", 18 );
				tidyBufAppend( &escaped_errbuf, "<h2>Output Listing</h2>\n", 24 );
				tidyBufAppend( &escaped_errbuf, "<p>Below is my tidied up version:</p>\n", 38 );
				tidyBufAppend( &escaped_errbuf, "<ul>\n<li><code>", 15 );

				tidyBufAppend( &escaped_errbuf, "0001 ", 5 );
				line = 1;
				
				while( !tidyBufEndOfInput( &tidyhtmlbuf ) ) {
					int ch = tidyBufGetByte( &tidyhtmlbuf );
					
					if (ch == '&') {
						tidyBufAppend( &escaped_errbuf, "&amp;", 5 );
					}else if (ch == '<') {
						tidyBufAppend( &escaped_errbuf, "&lt;", 4 );
					}else if (ch == '>') {
						tidyBufAppend( &escaped_errbuf, "&gt;", 4 );
					}else if (ch == '\n') {
						tidyBufAppend( &escaped_errbuf, "</code></li>\n<li><code>", 23 );
						line ++;
						if (line < 10000) {
							sprintf( strline, "%.4d ", line );
							tidyBufAppend( &escaped_errbuf, strline, 5 );
						}
					}else{
						tidyBufPutByte( &escaped_errbuf, ch );
					}
				}

				r->content_type = "application/xml; charset=utf-8"; /* Set XHTML Content-Type */
				
				ap_fputstrs(f->next, ctx->bb, 
					"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
					"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \n"
					"\"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n"
					"<html xmlns=\"http://www.w3.org/1999/xhtml\">\n"
					"<head>\n"
					"<title>mod_tidy Error Log</title>\n"
					"<meta name=\"publisher\" content=\"The mod_tidy Team\"/>\n"
					"<meta name=\"author\" content=\"Sierk Bornemann, bornemann@users.sourceforge.net\"/>\n"
					"<meta name=\"revision\" content=\"$Id: mod_tidy.c 111 2006-01-26 15:44:29Z bornemann $\"/>\n"
					"<link rel=\"shortcut icon\" type=\"image/x-icon\" "
					"href=\"data:image/x-icon;base64," /* base64 encoded embedded image data as defined in RFC 2397 */
					"AAABAAEAEBAQAAAAAAAoAQAAFgAAACgAAAAQAAAAIAAAAAEABAAAAAAAwAAAAAAAAAAAAAAA"
					"AAAAAAAAAAAAAAAA////ADm3/wDmimAA7cKuAIvP/wDN6f8A9+LaAOaoiwBlvP8A8fT2APPX"
					"ygDzu50AaM//APDNvQDlo4UAiIiIj0ERERGIiIiPQREREYiIiI9BERERiIiIj0ERERGIiIiP"
					"QREREfiPj/PxEWERiPg4SqEWXRHku7dDzSlpEbe+t8OC1VJhjPg+s4IZUaH/j4+PwaEREYiI"
					"iI9BERERiIiIj0ERERGIiIiPQREREYiIiI9BERERiIiIj0EREREAAAAAAAAAAAAAAAAAAAAA"
					"AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\"/>\n"
					"<style type=\"text/css\">\n"
					"html {\n"
					"    background-color: #ffffff;\n"
					"}\n\n"
					"body {\n"
					"    color: #11111a;\n"
					"    background-color: ffffff;\n"
					"    font-family: Helvetica, \"Trebuchet MS\", Arial, sans-serif;\n"
					"    font-size: 1em;\n"
					"    padding-left: 6%;\n"
					"    padding-right: 6%;\n"
					"    margin: 0 auto;\n"
					"}\n\n"
					"a:link {\n"
					"    color: #0066cc;\n"
					"    background: transparent;\n"
					"    text-decoration: none;\n"
					"}\n\n"
					"a:visited {\n"
					"    color: #003399;\n"
					"    background: transparent;\n"
					"    text-decoration: none;\n"
					"}\n\n"
					"a:hover, a:active {\n"
					"    color: #005a9c;\n"
					"    background: transparent;\n"
					"    text-decoration: underline;\n"
					"}\n\n"
					"#footer a {\n"
					"    color: #ffffff !important;\n"
					"    background: transparent;\n"
					"    text-decoration: underline;\n"
					"}\n\n"
					"h1 {\n"
					"    color: #053188;\n"
					"    background: transparent;\n"
					"    font-family: \"Bitstream Vera Sans\", Helvetica, Arial, sans-serif;\n"
					"    font-size-adjust: .53;\n"
					"    font-weight: bold;\n"
					"}\n\n"
					"h2 {\n"
					"    color: #777742;\n"
					"    background-color: #ffffff;\n"
					"    font-family: Georgia, Didot, serif;\n"
					"    font-size: 1.3em;\n"
					"    text-align: center;\n"
					"    padding: 5px 15px;\n"
					"    margin-top: 20px;\n"
					"    margin-bottom: 20px;\n"
					"}\n\n"
					"h2.invalid {\n"
					"    text-align: center;\n"
					"    font-family: \"Bitstream Vera Sans\", Helvetica, Arial ,sans-serif;\n"
					"    font-size-adjust: .53;\n"
					"    font-size: 1.5em;\n"
					"    font-weight: 500;\n"
					"    font-style: normal;\n"
					"    text-decoration: none;\n"
					"    word-spacing: normal;\n"
					"    letter-spacing: normal;\n"
					"    border-color: #000000;\n"
					"    padding-top: 0.5em;\n"
					"    padding-bottom: 0.5em;\n"
					"    margin-bottom: 0;\n"
					"    clear: right;\n"
					"}\n\n"
					"h3 {\n"
					"    color: #ffffff;\n"
					"    background-color: #aaaa77;\n"
					"    font-family: Georgia, Didot, serif;\n"
					"    font-size: 1.1em;\n"
					"    padding: 2px 15px;\n"
					"    border-top: 1px dotted #777742;\n"
					"    border-bottom: 1px dotted #777742;\n"
					"}\n\n"
					"div#banner {\n"
					"    color: #339900;\n"
					"    background-color: #8ca8e6;\n"
					"    padding: 0;\n"
					"    margin-bottom: 0;\n"
					"    height: 50px;\n"
					"}\n\n"
					"#banner h1 {\n"
					"    color: #ffffff;\n"
					"    background: transparent;\n"
					"    height: 50px;\n"
					"    vertical-align: middle;\n"
					"    font-size: 1.5em;\n"
					"    display: inline;\n"
					"}\n\n"
					"#banner h1 .alt {\n"
					"    color: #ff9900;\n"
					"    background: transparent;\n"
					"}\n\n"
					"#banner h1 .subheading {\n"
					"    font-size: 0.6em;\n"
					"}\n\n"
					"ul#menu {\n"
					"    background-color: #003399;\n"
					"    color: inherit;\n"
					"    margin: 0;\n"
					"    padding-top: 3px;\n"
					"    padding-right: 1em;\n"
					"    padding-bottom: 3px;\n"
					"    padding-left: 1em;\n"
					"    vertical-align: bottom;\n"
					"    font-size: 0.9em;\n"
					"    line-height: 140%;\n"
					"    list-style-type: none;\n"
					"}\n\n"
					"#menu li {\n"
					"    display: inline;\n"
					"}\n\n"
					"#menu li a {\n"
					"    color: #ffffff;\n"
					"    background: transparent;\n"
					"    padding-right: 1.2em;\n"
					"    padding-left: 1.2em;\n"
					"}\n\n"
					"#jumpbar {\n"
					"    color: inherit;\n"
					"    background-color: #dfeaff;\n"
					"    font-size: 0.9em;\n"
					"    font-weight: bold;\n"
					"    border-left: 1px solid #003399;\n"
					"    border-bottom: 1px solid #003399;\n"
					"    padding: 1ex;\n"
					"    margin: -15px 0 1em 1em;\n"
					"    width: 8em;\n"
					"    float: right;\n"
					"}\n\n"
					"#jumpbar ul {\n"
					"    padding: 0;\n"
					"    margin: 0;\n"
					"}\n\n"
					"#jumpbar li {\n"
					"    list-style-type: none;\n"
					"    text-align: right;\n"
					"}\n\n"
					"#jumpbar li a {\n"
					"    height: 20px;\n"
					"    padding-right: 20px;\n"
					"    padding-left: 25px;\n"
					"}\n\n"
					"#jumpbar li a:link, #jumpbar li a:visited {\n"
					"    border-top: 1px dotted #053188;\n"
					"    display: block;\n"
					"    width: 8em;\n"
					"    padding: 2px 0;\n"
					"    text-align: right;\n"
					"    text-decoration: none;\n"
					"}\n\n"
					"#jumpbar li a:hover {\n"
					"    text-decoration: underline;\n"
					"}\n\n"
					"p {\n"
					"    margin: 0.5em 25px 0 15px;\n"
					"}\n\n"
					"div.intro p {\n"
					"    font-style: italic;\n"
					"    text-align: justify;\n"
					"}\n\n"
					"div.intro p + p {\n"
					"    text-indent: 1.5em;\n"
					"    margin-top: 0;\n"
					"}\n\n"
					"p.warning {\n"
					"    color: inherit;\n"
					"    background-color: #c0c0c0;\n"
					"    font-size: small;\n"
					"    padding: 0.5em 25px 0.5em 15px;\n"
					"    margin: 0 0 20px 0;\n"
					"}\n\n"
					"div#main {\n"
					"    border-top 0;\n"
					"    border-right: 1px solid #dfeaff;\n"
					"    border-bottom: 0;\n"
					"    border-left: 1px solid #dfeaff;\n"
					"}\n\n"
					"#main, #footer {\n"
					"    padding: 15px 0;\n"
					"    margin: 0;\n"
					"}\n\n"
					"div#footer {\n"
					"    background-color: #6381dc;\n"
					"    color: #ffffff;\n"
					"    padding-top: 10px;\n"
					"    padding-bottom: 10px;\n"
					"    padding-left: 15px;\n"
					"    font-size: 90%;\n"
					"    clear: both;\n"
					"}\n\n"
					"#footer img {\n"
					"    border: 0;\n"
					"    display: block;\n"
					"    float: left;\n"
					"    margin-right: 1em;\n"
					"    margin-bottom: 2em;\n"
					"    width: 48px;\n"
					"    height: 15px;\n"
					"}\n\n"
					"#result ol,\n"
					"#source ul,\n"
					"#output ul {\n"
					"    list-style-type: none;\n"
					"}\n\n"
					"#result {\n"
					"    padding-top: 1em;\n"
					"}\n\n"
					"#result li.msg_err {\n"
					"    font-size: smaller;\n"
					"}\n\n"
					"#source ul,\n"
					"#output ul {\n"
					"    font-family: \"Bitstream Vera Sans Mono\", \"Lucida Console\", \"Andale Mono\", Courier, monospace;\n"
					"    color: inherit;\n"
					"    background-color: #eeeeee;\n"
					"    border: 1px solid #000000;\n"
					"    margin: 1em;\n"
					"    padding: 1em;\n"
					"}\n\n"
					"#source ul li,\n"
					"#output ul li {\n"
					"    margin: 0 0.5em;\n"
					"}\n\n"
					"code {\n"
					"    font-size: 0.9em;\n"
					"    white-space: pre;\n"
					"}\n\n"
					".invalid {\n"
					"    color: #ffffff;\n"
					"    background-color: #d23d24;\n"
					"}\n"
					"</style>\n"
					"</head>\n"
					"<body>\n"
					"<div id=\"banner\">\n"
					"<h1>\n"
					"mod_<span class=\"alt\">tidy</span>\n"
					"<span class=\"subheading\">Error Log</span>\n"
					"</h1>\n"
					"</div>\n"
					"<ul id=\"menu\">\n"
					"<li><a href=\"http://mod-tidy.sourceforge.net/\" title=\"The mod_tidy Project Page\"><strong>Home</strong></a></li>\n"
					"<li><a href=\"http://validator.w3.org/\" title=\"The W3C Markup Validation Service\">W3C Markup Validation Service</a></li>\n"
					"<li><a href=\"",TIDY_QUICKREF_URL,"\" title=\"Tidy Configuration Options (Quick Reference)\">Tidy Quick Reference</a></li>\n"
					"</ul>\n"
					"<div id=\"main\">\n"
					"<div id=\"jumpbar\">\n"
					"Jump to:\n"
					"<ul>\n"
					"<li><a href=\"#result\" title=\"Result of Parsing\">Results</a></li>\n"
					"<li><a href=\"#source\" title=\"Listing of unchanged Source Input\">Source Listing</a></li>\n"
					"<li><a href=\"#output\" title=\"Listing of mod_tidy's changed Output\">Output Listing</a></li>\n"
					"</ul>\n"
					"</div>\n"
					"<div class=\"intro\">\n"
					"<p>This is the error page of mod_tidy, a <a href=\"http://tidy.sourceforge.net/\" title=\"TidyLib, the core of mod_tidy\">TidyLib</a> based module "
					"for the Apache 2 HTTP Server to parse/syntax-check, clean up and pretty-print Web documents in formats like HTML and XHTML for conformance to W3C Recommendations "
					"and other standards.</p>\n"
					"</div>\n"
					"<div id=\"result\">\n"
					"<h2 class=\"invalid\">This page is <strong>not</strong> valid!</h2>\n"
					"<p class=\"warning\">As long as you run into this error page, the Markup of the current Web document isn't valid. "
					"To avoid this error page, meaning to be able to view your document as expected, "
					"<strong>you are strongly encouraged to correct its Markup</strong>.</p>\n"
					"<p>Below are the results of parsing this document with mod_tidy:</p>\n"
					"<ol>\n"
					"<li class=\"msg_err\">",escaped_errbuf.bp,"</code></li>\n"
					"</ul>\n"
					"</div>\n"
					"</div>\n"
					"<div id=\"footer\">\n"
					"<a href=\"http://mod-tidy.sourceforge.net/\">"
					"<img alt=\"Checked with mod_tidy\" title=\"This document is checked with mod_tidy\" "
					"src=\"data:image/png;base64," /* base64 encoded embedded image data as defined in RFC 2397 */
					"iVBORw0KGgoAAAANSUhEUgAAADEAAAAPCAMAAAB6MtfzAAAABGdBTUEAAK/INwWK6QAAABl0"
					"RVh0U29mdHdhcmUAQWRvYmUgSW1hZ2VSZWFkeXHJZTwAAAASUExURTOZAP+ZAGZmZv///4yo"
					"5gAAAGYTjM0AAAAGdFJOU///////ALO/pL8AAACuSURBVHjaYmAiEbACBBADEwsuwIwNMLEC"
					"BBDpOgACCIsOZhjNyAxCUAKuAyCACOhgxtQBEEBAHUAaiMHOgLgFqoORkRFsDYSC6mNiAAgg"
					"kA4WCEIQyK6COI0RZhEDA0AAkaCDAaIBIICI1AHCDAwgDawAAYSkA90fMPWMcB1AwMQKEEBY"
					"4wNLPDBCnAQMK4AAIi4GGSH+ZgCFLkAAkR7nAAFEug6AAGIgOe0CBBgA9goIxllrHREAAAAA"
					"SUVORK5CYIJ=\"/></a>\n"
					"<address>\n"
					"<a href=\"http://mod-tidy.sourceforge.net/feedback\" title=\"Send Feedback for mod_tidy\">The mod_tidy Team</a><br/>\n"
					"</address>\n"
					"$Date: 2006-01-26 16:44:29 +0100 (Thu, 26 Jan 2006) $\n"
					"</div>\n"
					"</body>\n"
					"</html>\n",
					NULL );
			}else{
				ap_fputstrs(f->next, ctx->bb, ctx->htmlbuf.bp, NULL );
			}
		
            APR_BUCKET_REMOVE(e);
            APR_BRIGADE_INSERT_TAIL(ctx->bb, e);
            ap_pass_brigade(f->next, ctx->bb);
			
			mod_tidy_g_MemPool = NULL;
			/* apr_thread_mutex_unlock( mod_tidy_g_MemMutex ); */
			
#ifdef MOD_TIDY_DEBUG
			printf( "<<<mod_tidy_filter( 0x%X, 0x%X ) eos\n", f, bb );
#endif /* #ifdef MOD_TIDY_DEBUG */
			
            return APR_SUCCESS;
        }
		/* if(APR_BUCKET_IS_FLUSH(e)) {
			APR_BUCKET_REMOVE(e);
            APR_BRIGADE_INSERT_TAIL(ctx->bb, e);
            ap_pass_brigade(f->next, ctx->bb);
			
			mod_tidy_g_MemAllocator = NULL;
			//apr_thread_mutex_unlock( mod_tidy_g_MemMutex );
			
#ifdef MOD_TIDY_DEBUG
			printf( "<<<mod_tidy_filter( 0x%X, 0x%X ) flush\n", f, bb );
#endif // #ifdef MOD_TIDY_DEBUG
			
            return APR_SUCCESS;
		} */
		apr_bucket_read(e, &str, &len, APR_NONBLOCK_READ);
		APR_BUCKET_REMOVE(e);

		if (str) {
			tidyBufAppend( &ctx->htmlbuf, (void*)str, len );
		}
    }
	
	mod_tidy_g_MemPool = NULL;
	/* apr_thread_mutex_unlock( mod_tidy_g_MemMutex ); */
	
#ifdef MOD_TIDY_DEBUG
	printf( "<<<mod_tidy_filter( 0x%X, 0x%X )\n", f, bb );
#endif /* #ifdef MOD_TIDY_DEBUG */
	
    return APR_SUCCESS;
}

static void mod_tidy_register_hook(apr_pool_t *p)
{
    ap_register_output_filter("TIDY", mod_tidy_filter, NULL, AP_FTYPE_CONTENT_SET);
}

static void *mod_tidy_create_dirconfig(apr_pool_t *p, char *dir)
{
	/* module config */
	mod_tidy_config* cfg;
	
	cfg = (mod_tidy_config*)apr_pcalloc(p, sizeof(mod_tidy_config));

	cfg->tidyoptions = apr_table_make(p, 4);

	return cfg;
}

static const char *set_mod_tidy_option(cmd_parms *parms,
  void *mconfig, const char *w1, const char *w2)
{
	mod_tidy_config* cfg = (mod_tidy_config*)mconfig;
	
#ifdef MOD_TIDY_DEBUG
	printf( "set_mod_tidy_option\n" );
	printf( "key == %s\n", w1 );
	printf( "value == %s\n", w2 );
#endif /* #ifdef MOD_TIDY_DEBUG */

	/* append option; w1==key; w2==value */
	apr_table_set( cfg->tidyoptions, w1, w2 );

	return NULL;
}

static const command_rec mod_tidy_directives[] =
{
	AP_INIT_TAKE2(
		"TidyOption",
		set_mod_tidy_option,
		NULL,
		ACCESS_CONF|OR_OPTIONS,
		"TidyOption name value see tidy libary documentation "\
		TIDY_QUICKREF_URL
	),
	{NULL}
};

module AP_MODULE_DECLARE_DATA tidy_module = 
{
    STANDARD20_MODULE_STUFF,
    mod_tidy_create_dirconfig,	/* create per-directory config structure */
    NULL,        		/* merge per-directory config structures */
    NULL,			/* create per-server config structure */
    NULL,			/* merge per-server config structures */
    mod_tidy_directives,	/* command apr_table_t */
    mod_tidy_register_hook	/* register hooks */
};
